/*
  ogmcat -- utility for concatenating OGG media files

  Written by Moritz Bunkus <moritz@bunkus.org>

  Distributed under the GPL
  see the file COPYING for details
  or visit http://www.gnu.org/copyleft/gpl.html
*/

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <ogg/ogg.h>
#include <vorbis/codec.h>

#define OGMSPLIT
#include "ogmmerge.h"
#include "ogmstreams.h"
#include "queue.h"
#include "vorbis_header_utils.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#define BLOCK_SIZE 4096

class cat_packetizer_c: public q_c {
  private:
    ogg_int64_t old_granulepos;
    char        stype;
    double      sample_rate;
  public:
    cat_packetizer_c(double nsample_rate, ogg_stream_state *ostate,
                       char nstype);
    virtual ~cat_packetizer_c();
    
    virtual stamp_t make_timestamp(ogg_int64_t granulepos);
    virtual void    process(ogg_packet *op);
    virtual void    produce_eos_packet() {};
    virtual void    produce_header_packets() {};
    virtual void    reset() {};
};

typedef struct stream_t {
  int                 serial;
  int                 eos, comment;
  char                stype;
  ogg_stream_state    instate;
  ogg_stream_state    outstate;
  ogg_packet          header_packet;

  stream_header       sth;
  vorbis_info         vi;
  
  ogg_int64_t         last_granulepos, this_granulepos;
  ogg_int64_t         granulepos;
  ogg_int64_t         packetno;
  ogg_int64_t         num_frames, cur_frame;
  double              sample_rate, granulepossub;
  vorbis_comment      vc;
  int                 vc_unpacked;
  ogmmerge_page_t    *mpage;
  cat_packetizer_c   *pzr;
  struct stream_t    *next;
} stream_t;

typedef struct source_t {
  char     *name;
  int       fd;
  off_t     size;
  stream_t *streams;
  source_t *next;
} source_t;

source_t       *sources = NULL;
int             numstreams = 0;

ogg_int64_t     bread = 0;
ogg_int64_t     bwritten = 0, bwritten_all = 0;

FILE           *fout;
char           *fout_name;
int             fout_num = 1;

int             verbose = 0;
int             frontend_mode = 0;

int             last_percent = 0;

/*
 * cat_packetizer class functions
 */

cat_packetizer_c::cat_packetizer_c(double nsample_rate,
                                       ogg_stream_state *ostate,
                                       char nstype) {
  old_granulepos = 0;
  stype = nstype;
  sample_rate = nsample_rate;
  memcpy(&os, ostate, sizeof(ogg_stream_state));
}

cat_packetizer_c::~cat_packetizer_c() {
}

stamp_t cat_packetizer_c::make_timestamp(ogg_int64_t granulepos) {
  stamp_t stamp;
  
#ifndef XIPHCORRECTINTERLEAVING  
  if (stype == 't')
    stamp = (stamp_t)((double)granulepos * (double)1000000.0 /
                      sample_rate);
  else
#endif
    stamp = (stamp_t)((double)old_granulepos * (double)1000000.0 /
                      sample_rate);
  old_granulepos = granulepos;
  
  return stamp;
}

void cat_packetizer_c::process(ogg_packet *op) {
  ogg_stream_packetin(&os, op);
  queue_pages();
}

/*
 * General helper functions, usage
 */

#define die(s) _die(s, __FILE__, __LINE__)
void _die(const char *s, const char *file, int line) {
  fprintf(stderr, "die @ %s/%d : %s\n", file, line, s);
  exit(1);
}

void usage(char *fname) {
  fprintf(stdout,
    "Usage: %s [options] -o out in1 [in2 [in3 ...]]\n\n"
    " options:\n"
    "  in1, in2 ...     Sources that will be concatenated.\n"
    "  -o, --output out Use 'out' as the base name. This is mandatory.\n"
    "  --frontend       Enable frontend mode. Progress output will be\n"
    "                   terminated by \\n instead of \\r.\n"
    "  -v, --verbose    Be verbose and show each OGG packet.\n"
    "                   Can be used twice to increase verbosity.\n"
    "  -h, --help       Show this help.\n"
    "  -V, --version    Show version information.\n", fname);
}

void print_progress(ogg_int64_t current, ogg_int64_t num, char *s) {
  if (frontend_mode)
    fprintf(stdout, "Processing %s %lld/%lld (%lld%%)\n", s,
            current, num, current * 100 / num);
  else
    fprintf(stdout, "Processing %s %lld/%lld (%lld%%)\r", s,
            current, num, current * 100 / num);
  fflush(stdout);
}

void copy_ogg_packet(ogg_packet *dst, ogg_packet *src) {
  memcpy(dst, src, sizeof(ogg_packet));
  dst->packet = (unsigned char *)malloc(src->bytes);
  if (dst->packet == NULL)
    die("malloc");
  memcpy(dst->packet, src->packet, src->bytes);
}

/*
 * File handling
 */

void write_page(ogg_page *page) {
  int ih, ib;
  
  if (fout == NULL)
    return;

  ih = fwrite(page->header, 1, page->header_len, fout);
  ib = fwrite(page->body, 1, page->body_len, fout);

  bwritten += ih + ib;
  bwritten_all += ih + ib;

  if (verbose > 1)
    fprintf(stdout, "(%s) %d + %d bytes written\n", __FILE__, ih, ib);
}

void flush_pages(stream_t *stream) {
  ogmmerge_page_t *mpage;
  ogg_page        *page;

  stream->pzr->flush_pages();
  if (stream->mpage == NULL)
    stream->mpage = stream->pzr->get_page();
  while (stream->mpage != NULL) {
    mpage = stream->mpage;
    page = mpage->og;
    write_page(page);
    free(page->header);
    free(page->body);
    free(page);
    free(mpage);
    stream->mpage = stream->pzr->get_page();
  }
}

int pages_available() {
  stream_t *stream;
  int       eos_only;
  
  eos_only = 1;
  stream = sources->streams;
  while (stream != NULL) {
    if (stream->mpage != NULL) {
      stream = stream->next;
      eos_only = 0;
      continue;
    }
    if (stream->pzr->page_available()) {
      stream->mpage = stream->pzr->get_page();
      stream = stream->next;
      eos_only = 0;
      continue;
    }
    if (!stream->eos)
      return 0;
    stream = stream->next;
  }
  
  return 1 - eos_only;
}

void write_winner_page() {
  stream_t *winner, *cur;
  
  winner = sources->streams;
  cur = winner->next;
  while (cur != NULL) {
    if (cur->mpage != NULL) {
      if (winner->mpage == NULL)
        winner = cur;
      else if (cur->mpage &&
               (cur->mpage->timestamp < winner->mpage->timestamp))
        winner = cur;
    }
    cur = cur->next;
  }
  if (winner->mpage != NULL) {
    write_page(winner->mpage->og);
    free(winner->mpage->og->header);
    free(winner->mpage->og->body);
    free(winner->mpage->og);
    free(winner->mpage);
    winner->mpage = NULL;
  }
}

void write_all_winner_pages() {
  while (pages_available())
    write_winner_page();
}
  
void flush_all_streams() {
  stream_t *s;
  
  s = sources->streams;
  while (s != NULL) {
    flush_pages(s);
    s = s->next;
  }
}

void produce_eos_packets() {
  stream_t   *s;
  ogg_packet  p;

  s = sources->streams;
  while (s != NULL) {
    if (!s->eos) {
      p.packet = (unsigned char *)"";
      p.bytes = 1;
      p.e_o_s = 1;
      p.b_o_s = 0;
      p.packetno = s->packetno++;
      p.granulepos = s->this_granulepos - (ogg_int64_t)s->granulepossub;
      s->pzr->process(&p);
      s->eos = 1;
    }
    s = s->next;
  }
  write_all_winner_pages();
}

void write_stream_headers() {
  stream_t *stream;
  
  stream = sources->streams;
  while (stream != NULL) {
    ogg_stream_init(&stream->outstate, stream->serial);
    stream->pzr = new cat_packetizer_c(stream->sample_rate,
                                       &stream->outstate,
                                       stream->stype);
    stream->eos = 0;
    stream->granulepos = 0;
    stream->pzr->process(&stream->header_packet);
    flush_pages(stream);
    stream = stream->next;
  }
}

long get_queued_bytes() {
  long      bytes;
  stream_t *stream;
  
  stream = sources->streams;
  bytes = 0;
  while (stream != NULL) {
    bytes += stream->pzr->get_queued_bytes();
    stream = stream->next;
  }
  
  return bytes;
}

/*
 * stream and source adding and finding
 */

void add_stream(source_t *source, stream_t *ndmx) {
  stream_t *cur = source->streams;
  
  if (cur == NULL) {
    source->streams = ndmx;
    ndmx->next = NULL;
  } else {
    while (cur->next != NULL)
      cur = cur->next;
    cur->next = ndmx;
    ndmx->next = NULL;
  }
}

stream_t *find_stream(int fserial) {
  stream_t *cur = sources->streams;
  
  while (cur != NULL) {
    if (cur->serial == fserial)
      break;
    cur = cur->next;
  }
  
  return cur;
}

void add_source(char *name, int fd, off_t size) {
  source_t *new_src, *cur;  
  
  new_src = (source_t *)malloc(sizeof(source_t));
  if (new_src == NULL)
    die("malloc");
  memset(new_src, 0, sizeof(source_t));
  
  new_src->name = strdup(name);
  if (new_src->name == NULL)
    die("strdup");
  new_src->fd = fd;
  new_src->size = size;
  
  if (sources == NULL)
    sources = new_src;
  else {
    cur = sources;
    while (cur->next != NULL)
      cur = cur->next;
    cur->next = new_src;
  }
}

/*
 * all-mighty processing functions
 */

void probe_all_sources() {
  ogg_sync_state  sync;
  ogg_page        page;
  ogg_packet      pack;
  vorbis_comment  vc;
  source_t       *src;
  stream_t       *stream;
  char           *buf;
  int             np, res;
  long            nread;
  
  src = sources;
  while (src != NULL) {
    ogg_sync_init(&sync);
    while (1) {
      np = ogg_sync_pageseek(&sync, &page);
      if (np < 0) {
        fprintf(stderr, "(%s) ogg_sync_pageseek failed for '%s'.\n", __FILE__,
                src->name);
        exit(1);
      }
      if (np == 0) {
        buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
        if (!buf) {
          fprintf(stderr, "(%s) ogg_sync_buffer failed for '%s'.\n", __FILE__,
                  src->name);
          exit(1);
        }
        if ((nread = read(src->fd, buf, BLOCK_SIZE)) <= 0)
          return;
        ogg_sync_wrote(&sync, nread);
        bread += nread;
        continue;
      }

      if (!ogg_page_bos(&page))
        break;
      stream = (stream_t *)malloc(sizeof(stream_t));
      if (stream == NULL)
        die("malloc");
      memset(stream, 0, sizeof(stream_t));
      stream->serial = ogg_page_serialno(&page);
      if (ogg_stream_init(&stream->instate, stream->serial)) {
        fprintf(stderr, "(%s) ogg_stream_init failed\n", __FILE__);
        exit(1);
      }
      add_stream(src, stream);
      ogg_stream_pagein(&stream->instate, &page);
      ogg_stream_packetout(&stream->instate, &pack);
      copy_ogg_packet(&stream->header_packet, &pack);
      if ((pack.bytes >= 7) &&
          !strncmp((char *)&pack.packet[1], "vorbis", 6)) {
        stream->stype = 'V';
        vorbis_info_init(&stream->vi);
        vorbis_comment_init(&vc);
        if ((res = vorbis_synthesis_headerin(&stream->vi, &vc, &pack)) >= 0)
          stream->sample_rate = stream->vi.rate;
        else {
          fprintf(stderr, "(%s) Vorbis audio stream indicated " \
                  "but no Vorbis stream header found in '%s'. Error code was "
                  "%d.\n", __FILE__, src->name, res);
          exit(1);
        }
      } else if (((*pack.packet & PACKET_TYPE_BITS ) == PACKET_TYPE_HEADER) &&
	         (pack.bytes >=
                   (int)(sizeof(stream_header) + 1 - sizeof(int)))) {
        stream_header *sth = (stream_header *)(pack.packet + 1);
        if (!strncmp(sth->streamtype, "video", 5)) {
          stream->sample_rate = (double)10000000 / (double)sth->time_unit;
          stream->stype = 'v';
        } else if (!strncmp(sth->streamtype, "audio", 5)) {
          stream->sample_rate = sth->samples_per_unit;
          stream->stype = 'a';
        } else if (!strncmp(sth->streamtype, "text", 4)) {
          stream->stype = 't';
          stream->sample_rate = (double)10000000 / (double)sth->time_unit;
        } else {
          fprintf(stderr, "(%s) Found new header of unknown/" \
                  "unsupported type in '%s'.\n", __FILE__, src->name);
          exit(1);
        }
        memcpy(&stream->sth, sth, sizeof(stream_header));
      } else {
        fprintf(stderr, "(%s) Found unknown header in '%s'.\n", __FILE__,
                src->name);
        exit(1);
      }
    }
    
    lseek(src->fd, 0, SEEK_SET);
    src = src->next;
  }
}

void check_all_sources() {
  source_t *src;
  stream_t *stream, *str;
  char      codec1[5], codec2[5];
  double    fps1, fps2;
  
  src = sources->next;
  while (src != NULL) {
    str = src->streams;
    if (str== NULL) {
      fprintf(stderr, "(%s) File '%s' does not contain any stream.\n",
              __FILE__, src->name);
      exit(1);
    }
    while (str != NULL) {
      stream = find_stream(str->serial);
      if (stream == NULL) {
        fprintf(stderr, "(%s) File '%s' contains a%s %s stream (serial %d) "
                "that is not present in '%s'.\n", __FILE__, src->name,
                str->stype == 'a' ? "n" : "",
                str->stype == 'V' ? "Vorbis" :
                  str->stype == 'v' ? "video" :
                  str->stype == 'a' ? "audio" :
                  str->stype == 't' ? "text" : "unknown??",
                str->serial, sources->name);
        exit(1);
      }
      if (str->stype != stream->stype) {
        fprintf(stderr, "(%s) File '%s' contains a%s %s stream (serial %d) "
                "that is of another type (%s) in '%s'.\n", __FILE__, src->name,
                str->stype == 'a' ? "n" : "",
                str->stype == 'V' ? "Vorbis" :
                  str->stype == 'v' ? "video" :
                  str->stype == 'a' ? "audio" :
                  str->stype == 't' ? "text" : "unknown??",
                str->serial,
                stream->stype == 'V' ? "Vorbis" :
                  stream->stype == 'v' ? "video" :
                  stream->stype == 'a' ? "audio" :
                  stream->stype == 't' ? "text" : "unknown??",
                sources->name);
        exit(1);
      }
      switch (str->stype) {
        case 'V':
          if (str->vi.rate != stream->vi.rate) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type Vorbis, rate %ld != %ld\n", __FILE__,
                    sources->name, src->name, str->serial,
                    str->vi.rate, stream->vi.rate);
            exit(1);
          }
          if (str->vi.channels != stream->vi.channels) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type Vorbis, channels %d != %d\n", __FILE__,
                    sources->name, src->name, str->serial,
                    str->vi.channels, stream->vi.channels);
            exit(1);
          }
          break;
        case 'v':
          codec1[4] = 0;
          codec2[4] = 0;
          memcpy(codec1, str->sth.subtype, 4);
          memcpy(codec2, stream->sth.subtype, 4);
          if (strcmp(codec1, codec2)) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type video, codec %s != %s\n", __FILE__,
                    sources->name, src->name, str->serial, codec1, codec2);
            exit(1);
          }
          
          fps1 = (double)10000000 / (double)str->sth.time_unit;
          fps2 = (double)10000000 / (double)stream->sth.time_unit;
          if (fps1 != fps2) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type video, fps %.3f != %.3f\n", __FILE__,
                    sources->name, src->name, str->serial, fps1, fps2);
            exit(1);
          }
          
          if (str->sth.sh.video.width != stream->sth.sh.video.width) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type video, width %d != %d\n", __FILE__,
                    sources->name, src->name, str->serial,
                    str->sth.sh.video.width, stream->sth.sh.video.width);
            exit(1);
          }
          
          if (str->sth.sh.video.height != stream->sth.sh.video.height) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type video, height %d != %d\n", __FILE__,
                    sources->name, src->name, str->serial,
                    str->sth.sh.video.height, stream->sth.sh.video.height);
            exit(1);
          }
          
          break;
        case 'a':
          codec1[4] = 0;
          codec2[4] = 0;
          memcpy(codec1, str->sth.subtype, 4);
          memcpy(codec2, stream->sth.subtype, 4);
          if (strcmp(codec1, codec2)) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type audio, codec %s != %s\n", __FILE__,
                    sources->name, src->name, str->serial, codec1, codec2);
            exit(1);
          }
          
          if (str->sth.samples_per_unit != stream->sth.samples_per_unit) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type audio, samples per second %lld != %lld\n",
                    __FILE__, sources->name, src->name, str->serial,
                    str->sth.samples_per_unit, stream->sth.samples_per_unit);
            exit(1);
          }

          if (str->sth.bits_per_sample != stream->sth.bits_per_sample) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type audio, bits per sample %d != %d\n",
                    __FILE__, sources->name, src->name, str->serial,
                    str->sth.bits_per_sample, stream->sth.bits_per_sample);
            exit(1);
          }

          if (str->sth.sh.audio.channels != stream->sth.sh.audio.channels) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type audio, channels %d != %d\n", __FILE__,
                    sources->name, src->name, str->serial,
                    str->sth.sh.audio.channels, stream->sth.sh.audio.channels);
            exit(1);
          }

          break;
        case 't':
          if (str->sth.time_unit != stream->sth.time_unit) {
            fprintf(stderr, "(%s) Stream parameter mismatch for '%s' and '%s':"
                    " serial %d, type text, time unit %lld != %lld\n",
                    __FILE__, sources->name, src->name, str->serial,
                    str->sth.time_unit, stream->sth.time_unit);
            exit(1);
          }

          break;
        default:
          fprintf(stderr, "(%s) Unsupported stream type in '%s'. This is a "
                  "bug in ogmcat. Please report it to moritz@bunkus.org.\n",
                  __FILE__, src->name);
      }
      str = str->next;
    }
  
    src = src->next;
  }
}

void process_ogms() {
/*  ogg_sync_state    sync;
  ogg_page          page;
  ogg_packet        pack;
  int               nread, np, sno, endofstream;
  stream_t         *stream;
  char             *buf;
  vorbis_info       vi;
  vorbis_comment    vc;
  int               hdrlen, i;
  long              lenbytes;
  splitpoint_t      sp;
  
  endofstream = 0;
  if (!print_mode && !cut_mode)
    next_sp = find_next_splitpoint(0);
  while (!endofstream) {
    sno = ogg_page_serialno(&page);
    stream = find_stream(sno);
    if (stream == NULL) {
      if (verbose > 1)
        fprintf(stdout, "(%s) Encountered packet for an unknown serial " \
                "%d !?\n", __FILE__, sno);
    } else {
      ogg_stream_pagein(&stream->instate, &page);
      stream->last_granulepos = stream->this_granulepos;
      stream->this_granulepos = ogg_page_granulepos(&page);
      while (ogg_stream_packetout(&stream->instate, &pack) == 1) {
        if (stream->stype != 'V') {
          if (stream->comment == 0) {
            if (print_mode) {
              copy_ogg_packet(&stream->header_packet2, &pack);
              if (stream->stype == 'v') {
                vorbis_unpack_comment(&stream->vc, (char *)pack.packet,
                                      pack.bytes);
                stream->vc_unpacked = 1;
              }
            }
            stream->comment = 1;
            stream->packetno = 2;
            continue;
          }
        } else {
          if (stream->comment == 0) {
            if (print_mode)
              copy_ogg_packet(&stream->header_packet2, &pack);
            stream->comment = 1;
            continue;
          } else if (stream->comment == 1) {
            if (print_mode)
              copy_ogg_packet(&stream->header_packet3, &pack);
            stream->comment = 2;
            stream->packetno = 3;
            continue;
          }
        }
        
        if (pack.granulepos == -1)
          pack.granulepos = stream->this_granulepos;
        current_time = (double)stream->last_granulepos * 1000.0 /
                       stream->sample_rate;
        if (cut_mode) {
          if ((current_cut != NULL) && (fout == NULL) &&
              (current_time == 0) && (current_cut->start == 0.0) &&
            !print_mode) {
            open_out_file();
            write_stream_headers();
          }
          if ((current_cut != NULL) && (current_time >= current_cut->end) &&
              (fout != NULL))
            do_close = 1;
        } else if ((fout == NULL) && !print_mode) {
          open_out_file();
          write_stream_headers();
          do_open = 0;
        }
        if ((stream->stype == 'v') && (pack.packet[0] & PACKET_IS_SYNCPOINT)) {
          stream->pzr->flush_pages();
          if (print_mode && (vstream == stream)) {
            ogg_int64_t pts = (ogg_int64_t)current_time;
            sp.pos_time = current_time;
            sp.pos_bytes = bwritten + get_queued_bytes();
            sp.frameno = stream->granulepos;
            add_splitpoint(&sp);
            if (print_mode & PM_PRINTSPLITPOINTS)
              fprintf(stdout, "(%s) Split point: %d, frameno: %lld, "
                      "bytes: %lld, start: %02d:%02d:%02d.%03d\n", __FILE__,
                      num_splitpoints - 1, stream->granulepos,
                      bwritten + get_queued_bytes(),
                      (int)(pts / 3600000),
                      (int)(pts / 60000) % 60,
                      (int)(pts / 1000) % 60,
                      (int)(pts % 1000));
          } else if (cut_mode) {
            if ((current_cut != NULL) && (fout == NULL) &&
                (current_time >= current_cut->start))
              do_open = 1;
          } else if (!print_mode &&
              (fout_num <= max_num_files) &&
              (next_sp != -1) &&
              (
               ((split_bytes > 0) &&
                ((bwritten_all + get_queued_bytes()) >=
                    splitpoints[next_sp].pos_bytes)
               ) ||
               (
                (split_time > 0) &&
                (current_time >= splitpoints[next_sp].pos_time)
               )
              )) {
            do_open = 1;
            do_close = 1;
          }
        }
        if (stream->stype == 'v') {
          if (do_close && !print_mode) {
            produce_eos_packets();
            if (fout != NULL) {
              fclose(fout);
              fout = NULL;
            }
            if (verbose) {
              ogg_int64_t sst = (ogg_int64_t)current_time;
              ogg_int64_t tf = (ogg_int64_t)(current_time - last_split);

              fprintf(stdout, "\n(%s) Closing file after writing %lld bytes "
                      "(%lld bytes written in all files), %02d:%02d:%02d.%03d "
                      "in this file (%02d:%02d:%02d.%03d elapsed).\n", __FILE__,
                      bwritten, bwritten_all,
                      (int)(tf / 3600000),
                      (int)(tf / 60000) % 60,
                      (int)(tf / 1000) % 60,
                      (int)(tf % 1000),
                      (int)(sst / 3600000),
                      (int)(sst / 60000) % 60,
                      (int)(sst / 1000) % 60,
                      (int)(sst % 1000));
            }
            if (cut_mode) {
              if (current_cut != NULL)
                current_cut = current_cut->next;
              if (current_cut == NULL)
                return;
            }
          }
          if (do_open && !print_mode) {
            stream_t *cur;

            if (!cut_mode)
              next_sp = find_next_splitpoint(next_sp);
            open_out_file();
            write_stream_headers();
            last_split = current_time;
            cur = str_first;
            while (cur != NULL) {
              cur->granulepossub = current_time * cur->sample_rate / 1000.0;
              cur = cur->next;
            }
          }
          do_open = 0;
          do_close = 0;


          hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
          hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
          lenbytes = 1;
          if ((hdrlen > 0) && (pack.bytes >= (hdrlen + 1)))
            for (i = 0, lenbytes = 0; i < hdrlen; i++) {
              lenbytes = lenbytes << 8;
              lenbytes += *((unsigned char *)pack.packet + hdrlen - i);
            }
          pack.granulepos = stream->granulepos + lenbytes;
          stream->granulepos += lenbytes;

          if (print_mode)
            stream->num_frames += lenbytes;
          else {
            stream->cur_frame += lenbytes;
            if ((verbose <= 1) &&
                (stream->cur_frame % (stream->num_frames / 100) == 0))
              print_progress(stream->cur_frame, stream->num_frames, "frame");
          }
        } else
          pack.granulepos -= (ogg_int64_t)stream->granulepossub;
        pack.packetno = stream->packetno++;
        if (pack.e_o_s)
          stream->eos = 1;
        if (!cut_mode ||
            ((current_cut != NULL) && (current_time >= current_cut->start))) {
          stream->pzr->process(&pack);
          if (stream->stype == 't')
            stream->pzr->flush_pages();
        } else {
          count_only = 1;
          stream->pzr->process(&pack);
          if (stream->stype == 't')
            stream->pzr->flush_pages();
        }
        write_all_winner_pages();
        count_only = 0;
      }
    }

    while (ogg_sync_pageseek(&sync, &page) <= 0) {
      buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
      nread = read(fdin, buf, BLOCK_SIZE);
      if (nread <= 0) {
        endofstream = 1;
        break;
      } else {
        ogg_sync_wrote(&sync, nread);
        bread += nread;
        if ((print_mode & PM_FINDSPLITPOINTS) && (verbose <= 1) &&
            (last_percent !=
              (int)((double)bread * 100.0 / (double)file_size))) {
          print_progress(bread, file_size, "bytes");
          last_percent = (int)((double)bread * 100.0 / (double)file_size);
        }
      }
    }
  }

  produce_eos_packets();  
  if (!print_mode) {
    fclose(fout);
    if (verbose) {
      ogg_int64_t sst = (ogg_int64_t)current_time;
      ogg_int64_t tf = (ogg_int64_t)(current_time - last_split);

      fprintf(stdout, "(%s) Closing file after writing %lld bytes "
              "(%lld bytes written in all files), %02d:%02d:%02d.%03d "
              "in this file (%02d:%02d:%02d.%03d elapsed).\n", __FILE__,
              bwritten, bwritten_all,
              (int)(tf / 3600000),
              (int)(tf / 60000) % 60,
              (int)(tf / 1000) % 60,
              (int)(tf % 1000),
              (int)(sst / 3600000),
              (int)(sst / 60000) % 60,
              (int)(sst / 1000) % 60,
              (int)(sst % 1000));
    }
  }
  
  stream = str_first;
  while (stream != NULL) {
    delete(stream->pzr);
    stream->mpage = NULL;
    stream = stream->next;
  }
  ogg_sync_clear(&sync);
  
  if ((verbose <= 1) && !print_mode) {
    print_progress(vstream->cur_frame, vstream->num_frames, "frame");
    if (!frontend_mode)
      fprintf(stdout, "\n");
  } else if ((verbose <= 1) && (print_mode & PM_FINDSPLITPOINTS)) {
    print_progress(file_size, file_size, "bytes");
    if (!frontend_mode)
      fprintf(stdout, "\n");
  }*/
}

int main(int argc, char *argv[]) {
  int   i, fdin = -1;
  off_t file_size;

  fprintf(stdout, "This program is not finished. Not even working. "
          "It's nothing. Ignore it for now.\n");

  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], "-V") || !strcmp(argv[i], "--version")) {
      fprintf(stdout, "ogmsplit v" VERSION "\n");
      exit(0);
    }
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--verbose"))
      verbose++;
    else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
      usage(argv[0]);
      return 0;
    } else if (!strcmp(argv[i], "--frontend"))
      frontend_mode = 1;
    else if (!strcmp(argv[i], "-o") || !strcmp(argv[i], "--output")) {
      if ((i + 1) > argc) {
        fprintf(stderr, "(%s) -o requires an argument.\n", __FILE__);
        exit(1);
      }
      fout_name = argv[i + 1];
      i++;
    } else {
      fdin = open(argv[i], O_RDONLY);
      if (fdin == -1) {
        fprintf(stderr, "(%s) Could not open \"%s\".\n", __FILE__,
                argv[i]);
        return 1;
      }
      file_size = lseek(fdin, 0, SEEK_END);
      if (file_size == (off_t)-1) {
        fprintf(stderr, "(%s) Could not seek to end of file.\n", __FILE__);
        return 1;
      }
      lseek(fdin, 0, SEEK_SET);
      add_source(argv[i], fdin, file_size);
    }
  }

  if ((sources == NULL) || (sources->next == NULL)) {
    fprintf(stdout, "(%s) Not enough source files given.\n", __FILE__);
    usage(argv[0]);
    exit(1);
  }
  if (fout_name == NULL) {
    fprintf(stdout, "(%s) No output file name given.\n", __FILE__);
    usage(argv[0]);
    exit(1);
  }
  
  probe_all_sources();
  check_all_sources();
  process_ogms();

  return 0;
}
